#include "asterx1_gps.h"

CasteRx1::CasteRx1(const string & hwPortName, const int acqRate)
{
	status = 0;
	config.portName = hwPortName;
	config.acqPeriod = acqPeriodValues[acqRate];
}

CasteRx1::~CasteRx1()
{
	blockList.clear();
	if( (status & ACQUISITION_RUNNING) ) stopAcquisition();
	if( (status & DEVICE_OPEN) ) closeDevice();
      delete serialPort;
	cout << "CasteRx1::~CasteRx1(). End of CasteRx1 destructor" << endl; //bye bye message
}

int CasteRx1::openDevice()
{
      TRS232_config serialConfig;
	int retValue = BASIC_SUCCESS;
	
	try{
		serialPort = new CRS232(config.portName); //creates a new object to manage serial connection
	}catch(CException &e)
	{
		cout << e.what() << endl;
		cout << "CasteRx1::openDevice(). ERROR. Allocation error when creating new object CRS232" << endl;
		retValue = PORT_OPEN_ERROR;
	}	
	
	if ( retValue == BASIC_SUCCESS ) //if successfully created , we can open it
	{
		try{
			serialPort->open( (void*)&config.portName ); 
		}catch(CCommException &e)
		{
			cout << e.what() << endl;
			cout << "CasteRx1::openDevice(). ERROR. Fail to open serial port" << endl;
			retValue = PORT_OPEN_ERROR;
		}
	}
	
	if ( retValue == BASIC_SUCCESS ) //if open is successful, we can configure
	{
		serialConfig.baud = ASTERX_BAUDRATE;
		serialConfig.num_bits = 8;
		serialConfig.parity = none;
		serialConfig.stop_bits = 1;
		try{
			serialPort->config( &serialConfig );
		}catch(CCommException &e)
		{
			cout << e.what() << endl;
			cout << "CasteRx1::openDevice(). ERROR. Fail to configure serial port" << endl;
			retValue = PORT_CONFIG_ERROR;
		}
	}
	
	if (retValue == BASIC_SUCCESS) 
	{
		status |= DEVICE_OPEN;
		cout << "CasteRx1::openDevice(). SUCCESS. Serial port " << config.portName << " opened successfully." << endl;
		
	}
	
	return retValue;	
}

int CasteRx1::closeDevice()
{
	int retValue = BASIC_SUCCESS;

	if ( (status & DEVICE_OPEN) ) 
	{
		try{
			serialPort->close();
			//status &= ~DEVICE_OPEN;
		}catch(CException &e)// catch(CCommException &e)
		{
			cout << e.what() << endl;
			cout << "CasteRx1::openDevice(). ERROR. Fail to close serial port " << config.portName << endl;
			retValue = PORT_CLOSE_ERROR;
			//status |= DEVICE_OPEN; //close() has failed , so it is still opened
		}
	}
	if ( retValue == BASIC_SUCCESS ) //if close() is successful, we delete the object
	{
		status &= ~DEVICE_OPEN;
		cout << "CasteRx1::closeDevice(). SUCCESS. Serial port " << config.portName << " closed successfully." << endl;		
	}
	
	return retValue;
}

int CasteRx1::startAcquisition()
{
	int retValue=0; 
	sbfBlockManager *blockMng;
	string cmnd;
	
	//enables commands at USB1 device input
	retValue += sendCommand("setDataInOut,USB1,CMD,none\n");

	//stops data transmission if it was running
	retValue += sendCommand("setSBFOutput,all,USB1,none\n");
	
	//enables septentrio binary format (SBF) as output at USB1
	retValue += sendCommand("setDataInOut,USB1,CMD,+SBF\n");
	
	//GPS signal processing options for urban mobile robots
	retValue += sendCommand("setMultipathMitigation,on\n");//enables processing for multipath removing (pag.26)
	retValue += sendCommand("setReceiverDynamics,Moderate\n");//Mobile robots moves at moderate speeds
	retValue += sendCommand("setPVTMode,Rover,StandAlone\n");//rover mode and enables standalone
	
	//data configuration. Add blocks of interest to "group 1"
	retValue += sendCommand("setSBFGroups,Group1,PVTCartesian\n");
	blockMng = new sbfBlockManager; blockMng->blockId = PVTCartesian_ID; blockMng->received = false;
	blockList.push_back(*blockMng);
	
	retValue += sendCommand("setSBFGroups,Group1,+PosCovCartesian\n");
	blockMng = new sbfBlockManager; blockMng->blockId = PosCovCartesian_ID; blockMng->received = false;
	blockList.push_back(*blockMng);

	retValue += sendCommand("setSBFGroups,Group1,+DOP\n");
	blockMng = new sbfBlockManager; blockMng->blockId = DOP_ID; blockMng->received = false;
	blockList.push_back(*blockMng);
	
	retValue += sendCommand("setSBFGroups,Group1,+PVTGeodetic\n");
	blockMng = new sbfBlockManager; blockMng->blockId = PVTGeodetic_ID; blockMng->received = false;
	blockList.push_back(*blockMng);
	
	//requests "Group1" data to be output to USB1 at specified rate 
	cmnd = "setSBFOutput,Stream1,USB1,Group1,";
	cmnd.append(config.acqPeriod);
	cmnd.append("\n");
	retValue += sendCommand(cmnd);
	
	if ( retValue == BASIC_SUCCESS*11 ) //success on all commands
	{
		status |= ACQUISITION_RUNNING;
		retValue = BASIC_SUCCESS;
		cout << "CasteRx1::startAcquisition(). SUCCESS on device configuration. The device is now transmitting data ..." << endl;
	}
	else //otherwise
	{
		status &= ~ACQUISITION_RUNNING;
		retValue = ASTERX1_CONFIG_ERROR;
		cout << "CasteRx1::startAcquisition(). ERROR on device configuration for data transmission" << endl;
	}

	return retValue;
}

int CasteRx1::stopAcquisition()
{
	int retValue=0; 
	
	if( (status & ACQUISITION_RUNNING) ) 
	{
		blockList.clear();
	
		retValue += sendCommand("setSBFOutput,Stream1,USB1,none\n");//stops data transmission from device
		retValue += sendCommand("setDataInOut,USB1,CMD,none\n");//enables commands as input at USB1		
		if ( retValue == BASIC_SUCCESS*2 )
		{
			status &= ~ACQUISITION_RUNNING;
			retValue = BASIC_SUCCESS;
			cout << "CasteRx1::stopAcquisition(). SUCCESS on stopping the device data transmission." << endl;
		}
		else 
		{
			retValue = ASTERX1_STOPPINTG_ERROR;
			cout << "CasteRx1::stopAcquisition(). ERROR on stopping the device data transmission." << endl;
		}
	}
	else 
	{
		cout << "CasteRx1::stopAcquisition(). Nothing to do, the acquisition was already stopped." << endl;
		retValue = BASIC_SUCCESS;
	}
	
	return retValue;
}

int CasteRx1::readDataFromDevice()
{
	int retValue;
	unsigned char buffer[MAX_BLOCK_SIZE];
	unsigned short int crcValue, crcComputed; //crc value from device and crc computed
	unsigned short int blockId; //data block id from receiver
	unsigned short int blockLength; //size of the SBF block in bytes
	timeval tmStmp;
	list<sbfBlockManager>::iterator iiBlock;	
	bool allReceived = false;
	
	//reset received field of block managers
	for (iiBlock=blockList.begin();iiBlock!=blockList.end();iiBlock++){ iiBlock->received = false; }
	
	//acquisition up to all requested blocks are received
	while (!allReceived)
	{
		//Sets Time Stamp (this is the system time stamp, not the GPS time!)
		gettimeofday(&tmStmp, NULL); 
		timeStamp = (double)(tmStmp.tv_sec + tmStmp.tv_usec/1e6);	
		
		retValue = synchronizeHeader((const unsigned char *)"$@", DATA_TIMEOUT); 
		if ( retValue != BASIC_SUCCESS ) //if !BASIC_SUCCESS, status will be either COMMTIMEOUT or SERIALCOMMERROR
		{
			status &= ~DATA_BLOCK_OK;
			cout << "CasteRx1::readDataFromDevice(). ERROR. when synchronizing with header. Expected header = $@" << endl;
			return ACQUISITION_ERROR; 
		}		

		//starts reading header
		retValue = 0;
		retValue += readNbytes(&buffer[0],2,DATA_TIMEOUT);//reads transmitted crc value
		crcValue = getUI2(&buffer[0]); //decodes crc value
		retValue += readNbytes(&buffer[0],2,DATA_TIMEOUT);//reads block id field
		blockId = (buffer[0] & 0xff)|((buffer[1] & 0x1f)<<8);//decodes blockId (uses only the 13 LS bits)
		retValue += readNbytes(&buffer[2],2,DATA_TIMEOUT);//reads block length
		blockLength = getUI2(&buffer[2]);//decodes block length (in bytes)

		//check for some errors
		if ( retValue != 3*BASIC_SUCCESS )
		{
			status &= ~DATA_BLOCK_OK;
			cout << "CasteRx1::readDataFromDevice(). ERROR. Error reading block header." << endl;
			return ACQUISITION_ERROR; 
		}
		if ( (blockLength%4!=0) || (blockLength < MIN_BLOCK_SIZE) || (blockLength > MAX_BLOCK_SIZE) )
		{
			status &= ~DATA_BLOCK_OK;
			cout << "CasteRx1::readDataFromDevice(). ERROR. SBF block Length not correct: " << blockLength << " bytes" << endl;
			return ACQUISITION_ERROR; 
		}
		
		//cout << "readDataFromDevice(): "<< __LINE__ << "; blockLength = " << blockLength << endl;
		
		//reads the rest of the block and loads it to the buffer
		retValue = readNbytes(&buffer[4],(blockLength-8),DATA_TIMEOUT);
		if ( retValue != BASIC_SUCCESS )
		{
			status &= ~DATA_BLOCK_OK;
			cout << "CasteRx1::readDataFromDevice(). ERROR. Error reading block data." << endl;
			return ACQUISITION_ERROR; 
		}

		//Computes & checks CRC
		crcComputed = CRC_compute16CCITT(buffer, blockLength-4);
		//cout << "Expected value: " << crcValue << "; Computed value: " << crcComputed << endl;
		if ( crcComputed != crcValue )
		{
			status &= ~DATA_BLOCK_OK;
			cout << "CasteRx1::readDataFromDevice(). ERROR. CRC error. Expected value: " << crcValue << "; Computed value: " << crcComputed << endl;
			return ACQUISITION_ERROR; 
		}

		//decodes SBF block
		switch(blockId)
		{
			case PVTCartesian_ID:
				retValue = decodePVTCartesian(&buffer[4]);
				markBlockAsReceived(PVTCartesian_ID);
				break;

			case PVTGeodetic_ID:
				retValue = decodePVTGeodetic(&buffer[4]);
				markBlockAsReceived(PVTGeodetic_ID);
				break;
				
			case PosCovCartesian_ID:
				retValue = decodePosCovCartesian(&buffer[4]);
				markBlockAsReceived(PosCovCartesian_ID);
				break;

			case DOP_ID:
				retValue = decodeDOP(&buffer[4]);
				markBlockAsReceived(DOP_ID);
				break;

			default:
				cout << "CasteRx1::readDataFromDevice(). ERROR. Unknown SBF blockId: " << blockId << endl;
				retValue = ACQUISITION_ERROR;
				break;
		}
	
		//check resulting status of the decoding function
		if ( retValue != BASIC_SUCCESS )
		{
			status &= ~DATA_BLOCK_OK;
			cout << "CasteRx1::readDataFromDevice(). ERROR on decoding SBF block. BlockId = " << blockId << ". Status = " << status << endl;
			return ACQUISITION_ERROR;
		}
		else//sets status and checks recursively if all blocks have been received
		{		
			status |= DATA_BLOCK_OK;
			allReceived = true;
			for (iiBlock=blockList.begin();iiBlock!=blockList.end();iiBlock++){ allReceived = iiBlock->received & allReceived; }
		}
	}

	//check gps availability
	if ( pvtError == 0 )
	{ 
		status |= GPS_AVAILABLE;
		fromWgsToMap(); //conversion to map coordinates
		computeTMenu(); //compute rotation matrix form WGS to local tangential ENU frame
		computeENUcovs(); //compute covariances from WGS  to local tangential ENU frame
	}
	else 
	{
		status &= ~GPS_AVAILABLE;
	}

	return BASIC_SUCCESS;
}

void CasteRx1::printData(ostream & ost)
{
	streamsize nn;

	ost << status << " ";
	nn=ost.precision(12); ost << timeStamp << " "; ost.precision(nn);
	ost << wnc << " " << tow << " " << numSatellites << " " << PDOP << " ";
	nn=ost.precision(12); ost << lat << " " << lon << " " << alt << " "; ost.precision(nn);
	ost << xWgs << " " << yWgs << " " << zWgs << " ";
	ost << xMap << " " << yMap << " " << zMap << " ";
	ost << vxWgs << " " << vyWgs << " " << vzWgs << " ";
	ost << vxMap << " " << vyMap << " " << vzMap << " ";
	ost << cxxWgs << " " << cyyWgs << " " << czzWgs << " " << cxyWgs << " ";
	ost << cxxMap << " " << cyyMap << " " << czzMap << " " << cxyMap << " ";
	ost << endl;
}

void CasteRx1::setMapOrigin(double mapLat0, double mapLon0, double mapAlt0, double mapAlpha0)
{
	lat0 = mapLat0*M_PI/180.0;
	lon0 = mapLon0*M_PI/180.0;	
	alt0 = mapAlt0;
	alpha0 = mapAlpha0*M_PI/180.0;
	
	computeTM();
}

void CasteRx1::setPortName(string pName)
{
	if( (status & DEVICE_OPEN) ) //if device is open, port name can't be changed
	{
		cout << "CasteRx1::setPortName(): WARNING! Serial port resource can't be changed when port is open. Keeping old settings: " << config.portName << endl;
	}
	else
	{
		config.portName = pName;
		cout << "CasteRx1::setPortName(): Serial port name setting has changed. New serial port is " << config.portName << endl;
	}
}

void CasteRx1::setAcquisitionRate(int rateIndex)
{
	if (status & ACQUISITION_RUNNING)
	{
		cout << "CasteRx1::setAcquisitionRate(): WARNING! Rate can't be changed when acquisition is running. Keeping old settings: " << config.acqPeriod << endl;
	}
	else
	{
		if ( ( rateIndex<ONCHANGE )|| ( rateIndex>SEC10 ) )
		{
			cout << "CasteRx1::setAcquisitionRate(): WARNING! Invalid rate index. Keeping old settings: " << config.acqPeriod << endl;
		}
		else
		{
			config.acqPeriod = acqPeriodValues[rateIndex];
			cout << "CasteRx1::setAcquisitionRate(): Acquisition rate has changed. New rate is " << config.acqPeriod << endl;
		}
	}
}

void CasteRx1::printTM()
{
	streamsize nn;
	
	cout << endl << "MAP ORIGIN (GEODETIC) ----------------------------------" << endl;
	nn=cout.precision(15);
	cout << "lat0 = " << lat0 << endl << "lon0 = " << lon0 << endl << "alt0 = " << alt0 << endl << "alpha0 = " << alpha0 << endl;
	cout.precision(nn);

	cout << endl << "MATRIX TRANSFORM FROM ECEF to MAP-----------------------" << endl;
	cout 	<< TM[0][0] << '\t' << TM[0][1] << '\t' << TM[0][2] << '\t' << TM[0][3] << endl
	<< TM[1][0] << '\t' << TM[1][1] << '\t' << TM[1][2] << '\t' << TM[1][3] << endl 
	<< TM[2][0] << '\t' << TM[2][1] << '\t' << TM[2][2] << '\t' << TM[2][3] << endl 
	<< TM[3][0] << '\t' << '\t' << TM[3][1] << '\t' << '\t' << TM[3][2] << '\t' << '\t' << TM[3][3] << endl;
	cout << "--------------------------------------------------------" << endl << endl;
}

unsigned int CasteRx1::getStatus(){ return status; }
double CasteRx1::getTimeStamp(){ return timeStamp; }
unsigned short int CasteRx1::getWnc(){ return wnc; }
unsigned int CasteRx1::getTow(){ return tow; }
unsigned int CasteRx1::getNumSatellites(){ return numSatellites; }
float CasteRx1::getPDOP(){ return PDOP; }
float CasteRx1::getTDOP(){ return TDOP; }
float CasteRx1::getHDOP(){ return HDOP; }
float CasteRx1::getVDOP(){ return VDOP; }
float CasteRx1::getUndulation(){ return undulation; }
unsigned short int CasteRx1::getPVTerror(){ return pvtError;}

double CasteRx1::getLat(bool units)
{ 
	if ( units == inRADS ) return lat; 
	else return lat*180.0/M_PI;
}

double CasteRx1::getLon(bool units)
{
	if ( units == inRADS ) return lon; 
	else return lon*180.0/M_PI;
}

double CasteRx1::getAlt(){ return alt; }
double CasteRx1::getXWgs(){ return xWgs; }
double CasteRx1::getYWgs(){ return yWgs; }
double CasteRx1::getZWgs(){ return zWgs; }
double CasteRx1::getXMap(){ return xMap; }
double CasteRx1::getYMap(){ return yMap; }
double CasteRx1::getZMap(){ return zMap; }
double CasteRx1::getVxWgs(){ return vxWgs; }
double CasteRx1::getVyWgs(){ return vyWgs; }
double CasteRx1::getVzWgs(){ return vzWgs; }
double CasteRx1::getVxMap(){ return vxMap; }
double CasteRx1::getVyMap(){ return vyMap; }
double CasteRx1::getVzMap(){ return vzMap; }
double CasteRx1::getCxxWgs(){ return cxxWgs; }
double CasteRx1::getCyyWgs(){ return cyyWgs; }
double CasteRx1::getCzzWgs(){ return czzWgs; }
double CasteRx1::getCxyWgs(){ return cxyWgs; }
double CasteRx1::getCxxMap(){ return cxxMap; }
double CasteRx1::getCyyMap(){ return cyyMap; }
double CasteRx1::getCzzMap(){ return czzMap; }
double CasteRx1::getCxyMap(){ return cxyMap; }
double CasteRx1::getCxxEnu(){ return cxxEnu; }
double CasteRx1::getCyyEnu(){ return cyyEnu; }
double CasteRx1::getCzzEnu(){ return czzEnu; }
double CasteRx1::getCxyEnu(){ return cxyEnu; }
double CasteRx1::getCxzEnu(){ return cxzEnu; }
double CasteRx1::getCyzEnu(){ return cyzEnu; }		


/******************************* PROTECTED MEMBERS ************************************************/

int CasteRx1::readNbytes(unsigned char *buffer, unsigned int nn, unsigned int msecTimeOut)
{
	CEventServer *eventServer=CEventServer::instance();
	string rxEvent;
	list<string> eventList;
	unsigned int bytesRead = 0;
	unsigned int queueSize = 0;
	unsigned int desiredBytes = 0;
	int retValue = BASIC_SUCCESS;
	
	//sets the event list with the reception event from serialPort object
	rxEvent = serialPort->get_rx_event_id();
	eventList.push_back(rxEvent);
	
	//first check if some byte is already in the queue
	queueSize = serialPort->get_num_data();
	if ( queueSize > 0)
	{
		desiredBytes = min( queueSize , (nn-bytesRead) );
		bytesRead += serialPort->read((unsigned char *)(buffer+bytesRead),desiredBytes);
	}

	//starts the waiting loop if not all bytes are already read
	while ( (bytesRead<nn) && (retValue == BASIC_SUCCESS) )
	{
		try{
			eventServer->wait_all(eventList, msecTimeOut);//waits for reception event
		}catch(CEventServerException &e)
		{
			cout << e.what() << ". Line: " << __LINE__ << endl;
			retValue = PORT_READ_TIMEOUT;
		}
		if (retValue == BASIC_SUCCESS)
		{
			queueSize = serialPort->get_num_data();
			desiredBytes = min( queueSize , (nn-bytesRead) );
			bytesRead += serialPort->read((unsigned char *)(buffer+bytesRead),desiredBytes);
		}
	}
	
	return retValue;
}

int CasteRx1::synchronizeHeader(const unsigned char *mark, int msecTimeOut, unsigned int markSize)
{
	unsigned char cc; 
	string readMark, targetMark;
	int retValue = BASIC_SUCCESS;
	unsigned int ii=0;

	//sets targetMark
	targetMark.append((const char*)mark,markSize);
	
	//reads mark from comm device
	for (ii=0; ii<markSize; ii++)
	{
		retValue = readNbytes(&cc,1,msecTimeOut);
		readMark.push_back((char)cc);
	}

	//continue reading while mark is not reached or while timeout
	while ( (readMark.compare(targetMark) != 0) && (retValue == BASIC_SUCCESS) )
	{
		for (ii=1; ii<markSize; ii++) readMark.at(ii-1) = readMark.at(ii); //left displacement of all bytes
		retValue = readNbytes(&cc,1,msecTimeOut); //gets a new byte
		readMark.at(markSize-1) = (char)cc; //replace the last byte by the new one
		//cout << endl << "readMark: " << readMark << endl << "targetMark: " << targetMark << endl;
	}

	return retValue;
}

int CasteRx1::sendCommand(const string & cmnd)
{
	int retValue;
	unsigned char replyBuffer[200];
	string receivedReply, expectedReply;
	
	//relax if we concatenate commands. (200 ms were suggested by Septentrio support team in a personal communication). 50ms tested and it seems it is not enough!!!
	usleep(100000);

	//sends the command string	
	serialPort->write((unsigned char*)cmnd.c_str(),cmnd.size());
	
	//synchronize with the reply header
	retValue = synchronizeHeader((const unsigned char *)"$R",500);
	
	//cout << "CasteRx1. Line: " << __LINE__ << ". Reply Sync achieved" << endl;
	if ( retValue != BASIC_SUCCESS )
	{
		cout << "CasteRx1::sendCommand(). ERROR. Synchronization with repky header $R not achieved after timeout" <<  endl;
		return retValue; //it will be either COMMTIMEOUT or SERIALCOMMERROR
	}

	//reads response up to receive expected length bytes (2 + cmnd.size()). Expected response : "$R: <cmnd>", but $R has been already read
	retValue = readNbytes(&replyBuffer[0],cmnd.size()+2,500);
	if ( retValue != BASIC_SUCCESS )
	{
		cout << "CasteRx1::sendCommand(). ERROR. Expected reply string not received after timeout" <<  endl;
		return retValue; //it will be either COMMTIMEOUT or SERIALCOMMERROR
	}
	
	//check response
	if ( replyBuffer[0] == '?' ) 
	{
		cout << "CasteRx1::sendCommand(). ERROR. Unknown command: " << cmnd << endl;
		return ASTERX1_INVALID_COMMAND;
	}
	receivedReply.append((const char*)replyBuffer,cmnd.size()+2);
	receivedReply.append("\n");
	expectedReply = ": ";
	expectedReply.append(cmnd);
	expectedReply.append("\n");
	if ( receivedReply.compare(0, cmnd.size()+1, expectedReply, 0, cmnd.size()+1) == 0) //expected = received
	{
		//cout << "CasteRx1::sendCommand(). Command Successful: " << cmnd << endl;
		return BASIC_SUCCESS;
	}
	else //otherwise
	{
		cout << "CasteRx1::sendCommand(). Error. " << endl;
		cout << "      Sent   command: " << cmnd << endl;
		cout << "      Device   Reply: " << receivedReply << endl;
		cout << "      Expected Reply: " << expectedReply << endl;
		return ASTERX1_CONFIG_ERROR;
	}
}

unsigned short int CasteRx1::getUI1(unsigned char *buf)
{
	//return ( (buf[0] & 0xff) | (0x0 << 8) );
	return (buf[0] & 0x00ff);
}

unsigned short int CasteRx1::getUI2(unsigned char *buf)
{
	return ( (buf[0] & 0xff) | ((buf[1]&0xff)<<8) );	
}

unsigned int CasteRx1::getUI4(unsigned char *buf)
{
	return ( (buf[0]&0xff) | ((buf[1]&0xff)<<8) | ((buf[2]&0xff)<<16) | ((buf[3]&0xff)<<24) );	
}

float CasteRx1::getFloat(unsigned char *buf)
{
	unsigned int uintL;
	uniFloat uniF;
	
	uintL=(buf[0]&0xff)|((buf[1]&0xff)<<8)|((buf[2]&0xff)<<16)|((buf[3]&0xff)<<24);
	uniF.i = uintL;
	
	return uniF.flt;	
}

double CasteRx1::getDouble(unsigned char *buf)
{
	unsigned int uintL;
	unsigned long long int uintH;
	uniDouble uniD;
	
	uintL=(buf[0]&0xff)|((buf[1]&0xff)<<8)|((buf[2]&0xff)<<16)|((buf[3]&0xff)<<24);
	uintH=(buf[4]&0xff)|((buf[5]&0xff)<<8)|((buf[6]&0xff)<<16)|((buf[7]&0xff)<<24);
	//uniD.i = uintL|((uintH&0xffff)<<32);
	uniD.i = uintL|uintH<<32;
	
	return uniD.dbl;	
}

int CasteRx1::decodePVTCartesian(unsigned char *sbfMsg)
{
	unsigned short int mode;//PVT computation mode (e.g. standAlone, fixed, differential, rtk,...)
	
	tow = getUI4(&sbfMsg[0]);
	wnc = getUI2(&sbfMsg[4]);
	mode = getUI1(&sbfMsg[6]);
	pvtError = getUI1(&sbfMsg[7]);
	if ( pvtError!=0 ) //GPS error to obtain position (this is not an acquisition error)
	{
		//cout << "CasteRx1::decodePVTCartesian(): PVT_ERROR: " << pvtError << endl;
	}
	else
	{
		xWgs = getDouble(&sbfMsg[8]);
		yWgs = getDouble(&sbfMsg[16]);	
		zWgs = getDouble(&sbfMsg[24]);
		undulation = (double) getFloat(&sbfMsg[32]);
		vxWgs = (double) getFloat(&sbfMsg[36]);
		vyWgs = (double) getFloat(&sbfMsg[40]);
		vzWgs = (double) getFloat(&sbfMsg[44]);
		numSatellites = getUI1(&sbfMsg[66]);
	}

	//cout << "CasteRx1::decodePVTCartesian()" << endl;
	return BASIC_SUCCESS;
}

int CasteRx1::decodePVTGeodetic(unsigned char *sbfMsg)
{
	unsigned short int mode;//PVT computation mode (e.g. standAlone, fixed, differential, rtk,...)
		
	tow = getUI4(&sbfMsg[0]);
	wnc = getUI2(&sbfMsg[4]);
	mode = getUI1(&sbfMsg[6]);
	pvtError = getUI1(&sbfMsg[7]);
	if ( pvtError!=0 ) //GPS error to obtain position (this is not an acquisition error)
	{
		//cout << "CasteRx1::decodePVTGeodetic(): PVT_ERROR: " << pvtError << endl;
	}
	else 
	{
		lat = getDouble(&sbfMsg[8]);
		lon = getDouble(&sbfMsg[16]);
		alt = getDouble(&sbfMsg[24]);
	}
	
	//cout << "CasteRx1::decodePVTGeodetic()" << endl;	
	return BASIC_SUCCESS;
}

int CasteRx1::decodePosCovCartesian(unsigned char *sbfMsg)
{	
	unsigned short int mode;//PVT computation mode (e.g. standAlone, fixed, differential, rtk,...)

	tow = getUI4(&sbfMsg[0]);
	wnc = getUI2(&sbfMsg[4]);
	mode = getUI1(&sbfMsg[6]);
	pvtError = getUI1(&sbfMsg[7]);
	if ( pvtError!=0 ) //GPS error to obtain position (this is not an acquisition error)
	{
		//cout << "CasteRx1::decodePosCovCartesian(): PVT_ERROR: " << pvtError << endl;
	}
	else 
	{
		cxxWgs = (double) getFloat(&sbfMsg[8]);
		cyyWgs = (double) getFloat(&sbfMsg[12]);
		czzWgs = (double) getFloat(&sbfMsg[16]);
		cxyWgs = (double) getFloat(&sbfMsg[24]);
		cxzWgs = (double) getFloat(&sbfMsg[28]);
		cyzWgs = (double) getFloat(&sbfMsg[36]);	
	}
	
	//cout << "CasteRx1::decodePosCovCartesian()" << endl;	
	return BASIC_SUCCESS;
}

int CasteRx1::decodeDOP(unsigned char *sbfMsg)
{
	unsigned char nrsv;//total number of satellites for DOP computation

	tow = getUI4(&sbfMsg[0]);
	wnc = getUI2(&sbfMsg[4]);
	nrsv = getUI1(&sbfMsg[6]);

	PDOP = (double)getUI2(&sbfMsg[8])/100.0;
	TDOP = (double)getUI2(&sbfMsg[10])/100.0;
	HDOP = (double)getUI2(&sbfMsg[12])/100.0;
	VDOP = (double)getUI2(&sbfMsg[14])/100.0;		

	//cout << "CasteRx1::decodeDOP()" << endl;
	return BASIC_SUCCESS;
}

//unsigned short CasteRx1::CRC_compute16CCITT(const void *buf, size_t buf_length)
unsigned short CasteRx1::CRC_compute16CCITT(unsigned char *buf, unsigned short int nBytes)
{
  unsigned int  ii;
  unsigned short  crc = 0;
  //const unsigned char  *buf8 = buf; /* Convert the type to access by byte. */

  /* see for example the BINEX web site */
  for (ii=0; ii<nBytes; ii++) {
    crc = (crc << 8) ^ CRC_16CCIT_LookUp[ (crc >> 8) ^ buf[ii] ];
  }

  return crc;
}

void CasteRx1::computeTM()
{
	double kk; //constant
	double xx0,yy0,zz0; //wgs-ecef coordinates of the origin
	double auxTM[4][4];
	
	//sets the rotation part of TM, from ECEF to ENU coordinates
	auxTM[0][0] = -sin(lon0);
	auxTM[0][1] = cos(lon0);
	auxTM[0][2] = 0.0;
	auxTM[1][0] = -sin(lat0)*cos(lon0);
	auxTM[1][1] = -sin(lat0)*sin(lon0);
	auxTM[1][2] = cos(lat0);
	auxTM[2][0] = cos(lat0)*cos(lon0);
	auxTM[2][1] = cos(lat0)*sin(lon0);
	auxTM[2][2] = sin(lat0);

	//computes final alignement of x with respect to the campus baseline as a rotation of alpha0 on z axis
	//Ra=[cos(alpha0)  sin(alpha0) 0;
	//   -sin(alpha0)  cos(alpha0) 0;
	//	0           0          1];
	TM[0][0]=cos(alpha0)*auxTM[0][0] + sin(alpha0)*auxTM[1][0];
	TM[0][1]=cos(alpha0)*auxTM[0][1] + sin(alpha0)*auxTM[1][1];
	TM[0][2]=cos(alpha0)*auxTM[0][2] + sin(alpha0)*auxTM[1][2];
	TM[1][0]=-sin(alpha0)*auxTM[0][0] + cos(alpha0)*auxTM[1][0];
	TM[1][1]=-sin(alpha0)*auxTM[0][1] + cos(alpha0)*auxTM[1][1];
	TM[1][2]=-sin(alpha0)*auxTM[0][2] + cos(alpha0)*auxTM[1][2];
	TM[2][0]=auxTM[2][0]; 
	TM[2][1]=auxTM[2][1];
	TM[2][2]=auxTM[2][2];

	//computes the translation part
	kk = EARTH_RADIUS/sqrt(1-(pow(EARTH_EXCENTRICITY*sin(lat0),2.0)));
	xx0 = (kk + alt0) * cos(lat0) * cos(lon0);
	yy0 = (kk + alt0) * cos(lat0) * sin(lon0);
	zz0 = ((1-EARTH_EXCENTRICITY*EARTH_EXCENTRICITY)* kk + alt0) * sin(lat0);
	TM[0][3] = - ( TM[0][0]*xx0 + TM[0][1]*yy0 + TM[0][2]*zz0 );
	TM[1][3] = - ( TM[1][0]*xx0 + TM[1][1]*yy0 + TM[1][2]*zz0 );
	TM[2][3] = - ( TM[2][0]*xx0 + TM[2][1]*yy0 + TM[2][2]*zz0 );
	TM[3][0] = 0.0; 
	TM[3][1] = 0.0; 
	TM[3][2] = 0.0;
	TM[3][3] = 1.0;
}

void CasteRx1::computeTMenu()
{
	//uses current position to compute ENU rotation 
	TMenu[0][0] = -sin(lon);
	TMenu[0][1] = cos(lon);
	TMenu[0][2] = 0.0;
	TMenu[1][0] = -sin(lat)*cos(lon);
	TMenu[1][1] = -sin(lat)*sin(lon);
	TMenu[1][2] = cos(lat);
	TMenu[2][0] = cos(lat)*cos(lon);
	TMenu[2][1] = cos(lat)*sin(lon);
	TMenu[2][2] = sin(lat);
}

void CasteRx1::fromWgsToMap()
{
	double auxM[3][3];
	int ii;

	//compute map coords
	xMap=TM[0][0]*xWgs+TM[0][1]*yWgs+TM[0][2]*zWgs+TM[0][3];
	yMap=TM[1][0]*xWgs+TM[1][1]*yWgs+TM[1][2]*zWgs+TM[1][3];
	zMap=TM[2][0]*xWgs+TM[2][1]*yWgs+TM[2][2]*zWgs+TM[2][3];

	//compute velocities 
	vxMap=TM[0][0]*vxWgs+TM[0][1]*vyWgs+TM[0][2]*vzWgs;
	vyMap=TM[1][0]*vxWgs+TM[1][1]*vyWgs+TM[1][2]*vzWgs;
	vzMap=TM[2][0]*vxWgs+TM[2][1]*vyWgs+TM[2][2]*vzWgs;

	//compute covariance matrix
	//firts step: auxM=TM*covEcef
	for(ii=0;ii<3;ii++)
	{
		auxM[ii][0]=TM[ii][0]*cxxWgs+TM[ii][1]*cxyWgs+TM[ii][2]*cxzWgs;
		auxM[ii][1]=TM[ii][0]*cxyWgs+TM[ii][1]*cyyWgs+TM[ii][2]*cyzWgs;
		auxM[ii][2]=TM[ii][0]*cxzWgs+TM[ii][1]*cyzWgs+TM[ii][2]*czzWgs;
	}
	//second step: covMap = auxM*TM^T
	cxxMap=auxM[0][0]*TM[0][0]+auxM[0][1]*TM[0][1]+auxM[0][2]*TM[0][2];
	cyyMap=auxM[1][0]*TM[1][0]+auxM[1][1]*TM[1][1]+auxM[1][2]*TM[1][2];
	czzMap=auxM[2][0]*TM[2][0]+auxM[2][1]*TM[2][1]+auxM[2][2]*TM[2][2];
	cxyMap=auxM[0][0]*TM[1][0]+auxM[0][1]*TM[1][1]+auxM[0][2]*TM[1][2];
}

void CasteRx1::computeENUcovs()
{
	double auxM[3][3];
	int ii;

	//compute covariance matrix
	//firts step: auxM=TMenu*covEcef
	for(ii=0;ii<3;ii++)
	{
		auxM[ii][0]=TMenu[ii][0]*cxxWgs+TMenu[ii][1]*cxyWgs+TMenu[ii][2]*cxzWgs;
		auxM[ii][1]=TMenu[ii][0]*cxyWgs+TMenu[ii][1]*cyyWgs+TMenu[ii][2]*cyzWgs;
		auxM[ii][2]=TMenu[ii][0]*cxzWgs+TMenu[ii][1]*cyzWgs+TMenu[ii][2]*czzWgs;
	}
	//second step: covMap = auxM*TMenu^T
	cxxEnu=auxM[0][0]*TMenu[0][0]+auxM[0][1]*TMenu[0][1]+auxM[0][2]*TMenu[0][2];
	cyyEnu=auxM[1][0]*TMenu[1][0]+auxM[1][1]*TMenu[1][1]+auxM[1][2]*TMenu[1][2];
	czzEnu=auxM[2][0]*TMenu[2][0]+auxM[2][1]*TMenu[2][1]+auxM[2][2]*TMenu[2][2];
	cxyEnu=auxM[0][0]*TMenu[1][0]+auxM[0][1]*TMenu[1][1]+auxM[0][2]*TMenu[1][2];
	cxzEnu=auxM[0][0]*TMenu[2][0]+auxM[0][1]*TMenu[2][1]+auxM[0][2]*TMenu[2][2];
	cyzEnu=auxM[1][0]*TMenu[2][0]+auxM[1][1]*TMenu[2][1]+auxM[1][2]*TMenu[2][2];
}

void CasteRx1::markBlockAsReceived(unsigned short int id)
{
	list<sbfBlockManager>::iterator iiBlock;	
	
	for (iiBlock=blockList.begin();iiBlock!=blockList.end();iiBlock++)
	{ 
		if ( iiBlock->blockId == id ) iiBlock->received = true;
	}
}
